"""
Constant values about endian.
"""

BIG_ENDIAN = "ABCD"
LITTLE_ENDIAN = "DCBA"
MIDDLE_ENDIAN = "BADC"
NETWORK_ENDIAN = BIG_ENDIAN

endian_name = {
    BIG_ENDIAN: "Big endian",
    LITTLE_ENDIAN: "Little endian",
    MIDDLE_ENDIAN: "Middle endian",
}
